"""aiozabbix - Asynchronous Zabbix API Python Library."""

#
# Original Ruby Library is Copyright (C) 2009 Andrew Nelson nelsonab(at)red-tux(dot)net
# Original Python Library is Copyright (C) 2009 Brett Lentz brett.lentz(at)gmail(dot)com
#
# Copyright (C) 2011-2018 Luke Cyca me(at)lukecyca(dot)com
# Copyright (C) 2018-2021 Modio AB
#
# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of the
# License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

from urllib.parse import urlparse
from typing import Optional, Dict, Any, Callable, Awaitable
import aiohttp
from aiohttp.client_exceptions import ContentTypeError
from opentelemetry import trace
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.util.types import Attributes

tracer = trace.get_tracer(__name__)

__all__ = [
    "ZabbixAPIException",
    "ZabbixAPI",
    "ZabbixAPIObjectClass",
]

# Type alias for Json-style data both to and from server.
JsonType = Any


class ZabbixAPIException(Exception):
    """Some kind of API exception."""


class ZabbixAPI:
    """Class representing the zabbix API."""

    # pylint: disable=too-many-instance-attributes

    DEFAULT_HEADERS = {"Content-Type": "application/json-rpc"}

    LOGIN_METHODS = ("user.login", "user.authenticate")
    UNAUTHENTICATED_METHODS = (
        "apiinfo.version",
        "user.checkAuthentication",
    ) + LOGIN_METHODS
    AUTH_ERROR_FRAGMENTS = (
        "authori",  # From CLocalApiClient::authenticate
        "permission",  # From CUser::checkAuthentication
        "re-login",  # From many places
    )

    def __init__(
        self,
        server: str = "http://localhost/zabbix",
        *,
        timeout: Optional[float] = None,
        client_session: Optional[aiohttp.ClientSession] = None,
        headers: Optional[dict[str, str]] = None,
    ):
        """Set up the zabbix api."""
        self.url = server + "/api_jsonrpc.php"

        if client_session is None:
            self.client_session = aiohttp.ClientSession()
        else:
            self.client_session = client_session

        self.timeout = None
        if timeout is not None:
            self.timeout = aiohttp.ClientTimeout(total=timeout)

        self.auth = ""
        self.shared_state = {"next_jsonrpc_id": 0}
        self.do_login: Optional[Callable[..., Awaitable[None]]] = None

        self.headers = self.DEFAULT_HEADERS.copy()
        if headers is not None:
            self.headers.update(headers)

    def with_headers(self, headers: Dict[str, str]) -> "ZabbixAPI":
        """Make a copy of the ZabbixAPI object which sets extra HTTP headers."""
        result = ZabbixAPI.__new__(ZabbixAPI)
        result.url = self.url
        result.client_session = self.client_session
        result.timeout = self.timeout
        result.auth = self.auth
        result.shared_state = self.shared_state
        result.do_login = self.do_login
        result.headers = self.headers.copy()
        result.headers.update(headers)
        return result

    async def do_request(
        self, method: str, params: Optional[JsonType] = None, auth_retries: int = 1
    ) -> JsonType:
        """Perform a request for the method with params."""
        rpc_id = self.shared_state["next_jsonrpc_id"]
        request_json = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params or {},
            "id": rpc_id,
        }
        attributes: Attributes = {
            SpanAttributes.NET_PEER_NAME: str(urlparse(self.url).hostname),
            SpanAttributes.RPC_SYSTEM: "jsonrpc",
            SpanAttributes.RPC_JSONRPC_VERSION: 2.0,
            SpanAttributes.RPC_METHOD: method,
            SpanAttributes.RPC_JSONRPC_REQUEST_ID: rpc_id,
        }

        with tracer.start_as_current_span("aiohttp.do_request", attributes=attributes):
            self.shared_state["next_jsonrpc_id"] += 1

            if method in self.UNAUTHENTICATED_METHODS:
                return await self.post_request(request_json)

            request_json["auth"] = self.auth

            try:
                return await self.post_request(request_json)
            except ZabbixAPIException as exc:
                if auth_retries > 0 and self.do_login and self.is_auth_error(exc):
                    await self.do_login(self)
                    return await self.do_request(
                        method, params, auth_retries=auth_retries - 1
                    )
                raise

    async def post_request(self, request_json: JsonType) -> JsonType:
        """Post a json request to the server."""
        response = await self.client_session.post(
            self.url, json=request_json, headers=self.headers, timeout=self.timeout
        )
        response.raise_for_status()

        try:
            response_json = await response.json()
        except (ValueError, ContentTypeError) as exc:
            text = await response.text()
            raise ZabbixAPIException(f"Unable to parse JSON: {text}") from exc

        if "error" in response_json:
            # Workaround for ZBX-9340, some errors don't contain 'data':
            if "data" not in response_json["error"]:
                response_json["error"]["data"] = "ZBX-9340: No data"

            err = response_json["error"]
            code = err["code"]
            msg = f"Error {code}: {err['message']}, {err['data']}"
            attributes = {
                SpanAttributes.RPC_JSONRPC_ERROR_CODE: code,
                SpanAttributes.RPC_JSONRPC_ERROR_MESSAGE: msg,
            }
            trace.get_current_span().set_attributes(attributes)
            raise ZabbixAPIException(msg, code)

        return response_json

    @classmethod
    def method_needs_auth(cls, method: str) -> bool:
        """Check if a method needs authentication."""
        return method not in cls.UNAUTHENTICATED_METHODS

    @classmethod
    def is_auth_error(cls, exc: Exception) -> bool:
        """Determine if an error is an authorization error.

        This makes a best effort attempt to recognize authentication
        or authorization errors. Unfortunately the general JSON-RPC
        error code -32602 (Invalid params) and the generic Zabbix
        error -32500 are used for these types of errors.

        The error messages may also be localized which could make this
        check fail.
        """

        err = str(exc).lower()
        return any(x in err for x in cls.AUTH_ERROR_FRAGMENTS)

    async def login(self, username: str = "", password: str = "") -> None:
        """Wrapper to log-in ot the remote server."""

        async def do_login(self: "ZabbixAPI") -> None:
            # Provide the self argument explicitly instead of taking
            # it from the surrounding closure. The self from the
            # closure will not be correct if this do_login is called
            # from a copy of the ZabbixAPI object created by the
            # with_headers method.

            # We create a new instance to not set the "auth" cookie to
            # something invalid while we are logging in.
            sub = self.with_headers({})
            sub.auth = ""
            try:
                # in zabbix 6.0+ one uses "username", so test if that works.
                self.auth = await sub.user.login(username=username, password=password)
            except ZabbixAPIException as ex:
                # otherwise, we use "user="  as well to test.
                try:
                    self.auth = await sub.user.login(user=username, password=password)
                except ZabbixAPIException as ex2:
                    # Tie the two exceptions together.
                    raise ex2 from ex

        self.do_login = do_login
        assert self.do_login is not None
        await self.do_login(self)

    async def confimport(
        self, confformat: str = "", source: str = "", rules: str = ""
    ) -> JsonType:
        """Compatibility function."""
        return await self.configuration.import_(
            format=confformat, source=source, rules=rules
        )

    def __getattr__(self, attr: str) -> "ZabbixAPIObjectClass":
        return ZabbixAPIObjectClass(name=attr, parent=self)


class ZabbixAPIObjectClass:
    """Method wrapper on the zabbix api."""

    # pylint: disable=too-few-public-methods

    def __init__(self, *, name: str, parent: ZabbixAPI) -> None:
        """Class instantiation for the method wrapper."""
        self.name = name
        self.parent = parent

    def __getattr__(self, attr: str) -> Callable[..., JsonType]:
        """Used to resolve methods dynamically."""
        if attr == "import_":
            attr = "import"
        method_name = f"{self.name}.{attr}"

        async def method(*args: JsonType, **kwargs: JsonType) -> JsonType:
            """JSON-rpc method wrapper for ZabbixApi."""
            if args and kwargs:
                raise TypeError(
                    "Method may be called with positional arguments or "
                    "keyword arguments, but not both at the same time"
                )
            response = await self.parent.do_request(method_name, args or kwargs)
            result = response["result"]
            return result

        return method



"""Testcases for aiozabbix."""

from typing import List
from dataclasses import dataclass

from aiohttp import web
import pytest

from aiozabbix import ZabbixAPI, ZabbixAPIException


async def mock_jsonrpc(request):
    """Mock RPC server that mostly replies with nonsense json."""
    request.app["state"].requests.append(request)

    request_json = await request.json()
    assert request_json["jsonrpc"] == "2.0"
    assert "params" in request_json
    assert "id" in request_json

    result = None

    method = request_json["method"]
    if ZabbixAPI.method_needs_auth(method):
        if not request.app["state"].logged_in:
            return web.json_response(
                {
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32602,
                        "message": "Invalid params.",
                        "data": "Not authorised.",
                    },
                    "id": request_json["id"],
                }
            )

    if method == "user.login":
        request.app["state"].logged_in = True
        result = f"mock auth token {request.app['state'].next_auth_token}"
        request.app["state"].next_auth_token += 1

    if result is None:
        result = f"mock response for {method}"

    return web.json_response(
        {
            "jsonrpc": "2.0",
            "result": result,
            "id": request_json["id"],
        }
    )


@dataclass
class AppState:
    """State for the test app."""

    requests: List
    logged_in: bool
    next_auth_token: int


@pytest.fixture(name="mock_server_app")
def fixture_mock_server_app():
    """Web server fixture."""
    app = web.Application()
    app.router.add_post("/api_jsonrpc.php", mock_jsonrpc)
    app["state"] = AppState(requests=[], logged_in=False, next_auth_token=1)
    return app


@pytest.fixture(name="client_session")
async def fixture_client_session(aiohttp_client, mock_server_app):
    """Client session fixture."""
    return await aiohttp_client(mock_server_app)


@pytest.fixture(name="zapi")
async def fixture_zapi(client_session, mock_server_app):
    """Zabbix API fixture."""
    assert mock_server_app
    return ZabbixAPI(server="", client_session=client_session)


async def test_unauthenticated_calls_should_work(mock_server_app, zapi):
    """Test that unauth calls work."""
    await zapi.apiinfo.version()

    requests = [await r.json() for r in mock_server_app["state"].requests]
    assert len(requests) == 1
    assert requests[0]["method"] == "apiinfo.version"


async def test_authenticated_mock_calls_should_fail_before_login(zapi):
    """Login should be required."""
    with pytest.raises(ZabbixAPIException):
        await zapi.hostgroup.get()


async def test_login_should_work(mock_server_app, zapi):
    """Login should work."""
    await zapi.login(username="Admin", password="zabbix")

    requests = [await r.json() for r in mock_server_app["state"].requests]
    assert len(requests) == 1
    assert requests[0]["method"] == "user.login"
    assert requests[0]["params"] == {"username": "Admin", "password": "zabbix"}


async def test_authenticated_mock_calls_should_succeed_after_login(zapi):
    """Login should work."""
    await zapi.login(username="Admin", password="zabbix")
    response = await zapi.hostgroup.get()
    assert "error" not in response


async def test_auth_token_should_be_sent(mock_server_app, zapi):
    """Login auth token should be passed on."""
    await zapi.login(username="Admin", password="zabbix")
    await zapi.hostgroup.get()

    requests = [await r.json() for r in mock_server_app["state"].requests]

    assert "auth" not in requests[0]
    assert requests[1]["auth"] == "mock auth token 1"


async def test_auth_error_should_cause_auto_relogin(mock_server_app, zapi):
    """Login error should be transparent."""
    await zapi.login(username="Admin", password="zabbix")
    await zapi.hostgroup.get()

    mock_server_app["state"].logged_in = False

    await zapi.hostgroup.get()

    requests = [await r.json() for r in mock_server_app["state"].requests]
    assert len(requests) == 5
    assert requests[0]["method"] == "user.login"
    assert requests[1]["method"] == "hostgroup.get"
    # Second hostgroup.get that fails due to being logged out
    assert requests[2]["method"] == "hostgroup.get"
    assert requests[2]["auth"] == "mock auth token 1"
    # Auto login
    assert requests[3]["method"] == "user.login"
    # Retry of hostgroup.get
    assert requests[4]["method"] == "hostgroup.get"
    assert requests[4]["auth"] == "mock auth token 2"


async def test_import_underscore_attr_should_be_rewritten(mock_server_app, zapi):
    """Method call with the underscore."""
    await zapi.login(username="Admin", password="zabbix")
    await zapi.confimport(
        confformat="xml", rules={}, source="<zabbix_export>...</zabbix_export>"
    )

    requests = [await r.json() for r in mock_server_app["state"].requests]
    assert len(requests) == 2
    assert requests[1]["method"] == "configuration.import"
    assert requests[1]["params"] == {
        "format": "xml",
        "rules": {},
        "source": "<zabbix_export>...</zabbix_export>",
    }


async def test_custom_headers_should_be_sent(mock_server_app, client_session):
    """Method call with custom headers."""
    zapi = ZabbixAPI(
        server="", client_session=client_session, headers={"User-Agent": "zabbixapp"}
    )

    await zapi.apiinfo.version()

    requests = mock_server_app["state"].requests
    assert len(requests) == 1
    assert requests[0].headers["User-Agent"] == "zabbixapp"


async def test_zabbix_api_copies_should_share_state_correctly(
    mock_server_app, client_session
):
    """Method call with shared state."""
    zapi = ZabbixAPI(
        server="", client_session=client_session, headers={"User-Agent": "zabbixapp"}
    )

    await zapi.apiinfo.version()

    zapi_with_extra_header = zapi.with_headers({"X-Extra-Header": "Yes"})
    await zapi_with_extra_header.apiinfo.version()

    await zapi.apiinfo.version()

    requests = mock_server_app["state"].requests
    assert len(requests) == 3

    assert requests[0].headers["User-Agent"] == "zabbixapp"
    assert "X-Extra-Header" not in requests[0].headers

    assert requests[1].headers["User-Agent"] == "zabbixapp"
    assert requests[1].headers["X-Extra-Header"] == "Yes"

    assert requests[2].headers["User-Agent"] == "zabbixapp"
    assert "X-Extra-Header" not in requests[2].headers

    request_ids = [(await r.json())["id"] for r in mock_server_app["state"].requests]
    assert request_ids == [0, 1, 2]


async def test_zabbix_api_object_copies_should_relogin_correctly(mock_server_app, zapi):
    """Method call with a copy should relogin."""
    await zapi.login(username="Admin", password="zabbix")
    await zapi.hostgroup.get()

    mock_server_app["state"].logged_in = False
    zapi_with_extra_header = zapi.with_headers({"X-Extra-Header": "Yes"})

    await zapi_with_extra_header.hostgroup.get()

    requests = [await r.json() for r in mock_server_app["state"].requests]
    assert len(requests) == 5
    assert requests[0]["method"] == "user.login"
    assert requests[1]["method"] == "hostgroup.get"
    # Second hostgroup.get that fails due to being logged out
    assert requests[2]["method"] == "hostgroup.get"
    assert requests[2]["auth"] == "mock auth token 1"
    # Auto login
    assert requests[3]["method"] == "user.login"
    # Retry of hostgroup.get
    assert requests[4]["method"] == "hostgroup.get"
    assert requests[4]["auth"] == "mock auth token 2"


async def test_zabbix_api_args_and_kwargs(mock_server_app, zapi):
    """Method call arg and kwarg mix."""
    await zapi.login(username="Admin", password="zabbix")
    # Should work with args
    await zapi.sub.method("arg1", "arg2")
    # should work with kwargs
    await zapi.sub.method(kwarg1="kwarg", kwarg2="kwargier")
    # login + two requests should be performed
    assert len(mock_server_app["state"].requests) == 3
    # should fail with mix of args + kwargs
    with pytest.raises(TypeError):
        await zapi.sub.method("arg", kwarg1="kwarg", kwarg2="kwargier")
    # no request should have gone to the server.
    assert len(mock_server_app["state"].requests) == 3


async def error_not_json(request):
    """Simple server that returns non-json data."""
    request.app["state"].requests.append(request)
    await request.json()
    return web.Response(
        text="This is not JSON?  Why would a server send something other than JSON?"
    )


async def test_odd_http_error_should_raise(aiohttp_client):
    """Set up a web server that returns HTTP data rather than JSON."""
    app = web.Application()
    app.router.add_post("/api_jsonrpc.php", error_not_json)
    app["state"] = AppState(requests=[], logged_in=False, next_auth_token=1)
    client_session = await aiohttp_client(app)
    zapi = ZabbixAPI(server="", client_session=client_session)

    with pytest.raises(ZabbixAPIException):
        await zapi.apiinfo.version()


async def error_old_login(request):
    """Simple server that mimicks Zabbix 5.0 old-style login."""
    request.app["state"].requests.append(request)
    request_json = await request.json()
    method = request_json["method"]
    if method == "user.login":
        if "user" not in request_json["params"]:
            return web.json_response(
                {
                    "jsonrpc": "2.0",
                    "error": {
                        "code": -32602,
                        "message": "Invalid params.",
                        "data": "User absent? What is user?",
                    },
                    "id": request_json["id"],
                }
            )
        request.app["state"].logged_in = True
        result = f"Old server mock auth token {request.app['state'].next_auth_token}"
        request.app["state"].next_auth_token += 1

    result = f"mock response for {method}"
    return web.json_response(
        {
            "jsonrpc": "2.0",
            "result": result,
            "id": request_json["id"],
        }
    )


async def test_login_new_to_old(aiohttp_client):
    """Test with a mocked old-style zabbix-server that requires "user" rather
    than "username"."""
    app = web.Application()
    app.router.add_post("/api_jsonrpc.php", error_old_login)
    app["state"] = AppState(requests=[], logged_in=False, next_auth_token=1)
    client_session = await aiohttp_client(app)
    zapi = ZabbixAPI(server="", client_session=client_session)

    await zapi.login(username="Admin", password="zabbix")
    requests = [await r.json() for r in app["state"].requests]
    assert len(requests) == 2
    assert requests[0]["method"] == "user.login"
    assert requests[0]["params"] == {"username": "Admin", "password": "zabbix"}
    assert requests[1]["method"] == "user.login"
    assert requests[1]["params"] == {"user": "Admin", "password": "zabbix"}


async def test_timeout_value_should_function(client_session):
    """Test that we can set the timeout."""
    zapi = ZabbixAPI(server="", client_session=client_session, timeout=0.1)
    await zapi.apiinfo.version()


async def test_default_session():
    """Test that it initializes without a session."""
    ZabbixAPI(server="")


async def error_all_logins(request):
    """Simple server that always errors on login."""
    request.app["state"].requests.append(request)
    request_json = await request.json()
    method = request_json["method"]
    if method == "user.login":
        return web.json_response(
            {
                "jsonrpc": "2.0",
                "error": {
                    "code": -32602,
                    "message": "Invalid params.",
                    "data": "User absent? What is user?",
                },
                "id": request_json["id"],
            }
        )

    result = f"mock response for {method}"
    return web.json_response(
        {
            "jsonrpc": "2.0",
            "result": result,
            "id": request_json["id"],
        }
    )


async def test_login_always_fails(aiohttp_client):
    """Test when login always fails."""
    app = web.Application()
    app.router.add_post("/api_jsonrpc.php", error_all_logins)
    app["state"] = AppState(requests=[], logged_in=False, next_auth_token=1)
    client_session = await aiohttp_client(app)
    zapi = ZabbixAPI(server="", client_session=client_session)

    with pytest.raises(ZabbixAPIException):
        await zapi.login(username="Admin", password="zabbix")

    # We should have made two attempts, and they should fail.
    requests = [await r.json() for r in app["state"].requests]
    assert len(requests) == 2
    assert requests[0]["method"] == "user.login"
    assert requests[0]["params"] == {"username": "Admin", "password": "zabbix"}
    assert requests[1]["method"] == "user.login"
    assert requests[1]["params"] == {"user": "Admin", "password": "zabbix"}
